import luigi
from luigi.contrib.external_program import ExternalProgramTask
from os import path
from wespipeline import utils 

class FastqcQualityCheck(ExternalProgramTask):
    """Task used for creating a quality report on fastq files.

    The report is created using the Fastqc utility, reulsting on an
    html report, an a zip folder containing more detailed information
    about the quality of the reads.

    Parameters:
        fastq_file (str): Path for the fastq file to be analyzed.

    Output:
        html (luigi.LocalTarget) : File containing the report for fastqc quality.

    """

    fastq_file = luigi.Parameter()

    def output(self):
        return {
            'html' : luigi.LocalTarget(self.fastq_file.replace('.fastq', '_fastqc.html')),
            'zip' : luigi.LocalTarget(self.fastq_file.replace('.fastq', '_fastqc.zip')),
        }

    def program_args(self):
        return ['fastqc', '-f', 'fastq', self.fastq_file]
    
class SraToolkitFastq(ExternalProgramTask):
    """Task used for downloading fastq files from the NVBI archive.

    In case of the reads to be paired end, the output will consist
    of two separate fastq files.

    The output file(s) will have for name the accession number and,
        in the case of paired end reads, a suffix identifying each
        of the two fastq.

    Parameters:
        accession_number (str): NCBI accession number for the experiment.
        paired_end (bool): Non case sensitive boolean indicating wether
            the reads are paired_end.

    Output:
        A dict mapping keys to `luigi.LocalTarget` instances for each of the 
        processed files. 
        The following keys are available:

        'fastq1' : Local file with the fastq file with the experiment's reads.     
        'fastq2' : In case of paired end experiments, a local file with the fastq 
            file with the experiment's reads.     

    """

    accession_number = luigi.Parameter()
    paired_end = luigi.Parameter()

    def output(self):
        output_files = {'fastq1' : luigi.LocalTarget(self.accession_number + '.fastq')}
        if self.paired_end.lower() == 'true':
            output_files = {'fastq1' : luigi.LocalTarget(self.accession_number + '_1.fastq'),
                'fastq2' : luigi.LocalTarget(self.accession_number + '_2.fastq')}

        return output_files

    def program_args(self):
        command = ['fastq-dump']

        if self.paired_end.lower() == 'true':
            command.append('--split-files')

        command.append(self.accession_number)

        return command

class UncompressFastqgz(luigi.Task):
    """Task for uncompressing fastq files.

    The task uses utils.UncompressFile for uncompressing into fastq.
    If both fastq_local_file and fastq_url are set, the local file will
    have preference; thus reducing the overhead in the process.

    Parameters:
        fastq_local_file (str): String indicating the location of a local
            compressed fastq file.
        fastq_url (str): Url indicating the location of the resource for
            the compressed fastq file.
        output_file (str): String indicating the desired location and name 
            the output uncompressed fastq file.

    Output:
        A `luigi.LocalTarget` instance for the uncompressed fastq file.

    """

    fastq_local_file = luigi.Parameter(default='')
    fastq_url = luigi.Parameter(default='')
    output_file = luigi.Parameter(default='')

    def requires(self):
        if self.fastq_local_file != '':
            return utils.LocalFile(file=self.fastq_local_file)
        else:
            return utils.Wget(url=self.fastq_url, output_file=self.output_file + '.gz')

    def output(self):
        return luigi.LocalTarget(self.output_file)

    def run(self):
        yield utils.UncompressFile(input_file=self.input().path, output_file=self.output().path)

class GetFastq(utils.MetaOutputHandler, luigi.Task):
    """Higher level task for the retrieval of the experiment fastq files.

    Three diferent sources for the fastq files are accepted: an existing local
    file, an NCBI accession number for the reads, and an external url indicating
    the location for the resources. The order in which the sources will be searched
    is the same as above: it is given preference to local files over external
    resources in order to reduce computational overhead, and NCBI accession number
    over external resources for reproducibility reasons.

    Parameters:
        fastq1_local_file (str): String indicating the location of a local
            compressed fastq file.
        fastq2_local_file (str): String indicating the location of a local
            compressed fastq file.
        fastq1_url (str): Url indicating the location of the resource for
            the compressed fastq file.
        fastq2_url (str): Url indicating the location of the resource for
            the compressed fastq file.
        paired_end (bool): Non case sensitive boolean indicating wether
            the reads are paired_end.
        compressed (bool): Non case sensitive boolean indicating wether
            the reads are compressed.
        create_report (bool): A non case-sensitive boolean indicating wether 
            to create a quality check report.

    Output:
        A dict mapping keys to `luigi.LocalTarget` instances for each of the 
        processed files. 
        The following keys are available:

        'fastq1' : Local file with the fastq file with the experiment's reads.     
        'fastq2' : In case of paired end experiments, a local file with the fastq 
            file with the experiment's reads.     

    """

    fastq1_local_file = luigi.Parameter(default='',description='Path to the fastq/ fastq.gz file. Use this parameter only if "from_local_file" is set to True.')
    fastq2_local_file = luigi.Parameter(default='',description='Path to the fastq/ fastq.gz file. Only used if the experiment is paired end.')

    fastq1_url = luigi.Parameter(default='', description="Url for the download of the sequenced data.")
    fastq2_url = luigi.Parameter(default='', description="Url for the download of the sequenced data. This will only be used for paired-end experiments.")

    accession_number = luigi.Parameter(default='', description="Optional string indicating the EBI accession number for retrieving the experiment.")

    paired_end = luigi.Parameter(default='false', description="A non case-sensitive boolean indicating wether the experimwnt is paired ended or not.")
    compressed = luigi.Parameter(default='false', description="A non case-sensitive boolean indicating wether the provided files are compressed or not.")
    create_report = luigi.Parameter(default='false', description="A non case-sensitive boolean indicating wether to create a quality check report.")

    def requires(self):
        if self.accession_number != '' and self.fastq1_local_file == '' and self.fastq1_local_file == '':
            return SraToolkitFastq(accession_number=self.accession_number, paired_end=self.paired_end)

        if self.compressed.lower() == 'true':
            if self.fastq1_local_file != '':
                dependencies = {'fastq1' : UncompressFastqgz(
                        fastq_local_file=self.fastq1_local_file, 
                        fastq_url=self.fastq1_url, 
                        output_file='.'.join(self.fastq1_local_file.split('.')[:-1]))}
            else: 
                dependencies = {'fastq1' : UncompressFastqgz(
                        fastq_local_file=self.fastq1_local_file, 
                        fastq_url=self.fastq1_url, 
                        output_file=path.join(utils.GlobalParams().base_dir, 'hg19_1.fastq'))}
        else:
            if self.fastq1_local_file != '':
                dependencies = {'fastq1' : utils.LocalFile(file=self.fastq1_local_file)}
            else:
                dependencies = {'fastq1' : utils.Wget(url=self.fastq1_url, output_file=path.join(utils.GlobalParams().base_dir, 'hg19_1.fastq'))}

        if self.paired_end.lower() == 'true':
            if self.compressed.lower() == 'true':
                if self.fastq2_local_file != '':
                    dependencies = {'fastq2' : UncompressFastqgz(
                            fastq_local_file=self.fastq2_local_file, 
                            fastq_url=self.fastq2_url, 
                            output_file='.'.join(self.fastq2_local_file.split('.')[:-1]))}
                else: 
                    dependencies = {'fastq2' : UncompressFastqgz(
                            fastq_local_file=self.fastq2_local_file, 
                            fastq_url=self.fastq2_url, 
                            output_file=path.join(utils.GlobalParams().base_dir, 'hg19_1.fastq'))}
            else:
                if self.fastq2_local_file != '':
                    dependencies = {'fastq2' : utils.LocalFile(file=self.fastq2_local_file)}
                else:
                    dependencies = {'fastq2' : utils.Wget(url=self.fastq2_url, output_file=path.join(utils.GlobalParams().base_dir, 'hg19_1.fastq'))}

        return dependencies

    def run(self):
        if self.create_report.lower() == 'true':
            report = [FastqcQualityCheck(fastq_file=self.input()['fastq1'].path)]
            if self.paired_end.lower() == 'true':
                report.append(FastqcQualityCheck(fastq_file=self.input()['fastq2'].path))

            yield report

if __name__ == '__main__':
    luigi.run(['GetFastq', 
            '--GetFastq-fastq1-url', '',
            '--GetFastq-fastq2-url', '',
            '--GetFastq-from-ebi', 'False',
            '--GetFastq-paired-end', 'False',
            '--GlobalParams-base-dir', path.abspath('./experiment'),
            '--GlobalParams-log-dir', path.abspath(path.curdir),
            '--GlobalParams-exp-name', 'hg19'])



import os
import luigi
from luigi.contrib.external_program import ExternalProgramTask
from wespipeline import utils 

class GetProgram(ExternalProgramTask):
    """Task user for downloading and giving execution permissions to the 2bit program.

    The task gives execute permissions to the conversion utility for 2bit files
    to be converted to fa files which can then be used for aligning the sequences.

    The source for the program is ftp://hgdownload.cse.ucsc.edu/admin/exe/linux.x86_64/twoBitToFa.

    Parameters:
        none

    Output:
        A `luigi.LocalTarget` for the executable.

    """
    
    def requires(self):
        program_file = path.join(utils.utils.GlobalParams().base_dir, 'twoBitToFa')
        program_url = 'ftp://hgdownload.cse.ucsc.edu/admin/exe/linux.x86_64/twoBitToFa'

        return Wget(url=program_url, output_file=program_file)

    def output(self):
        program_file = path.join(utils.utils.GlobalParams().base_dir, 'twoBitToFa')
        return luigi.LocalTarget(program_file)

    def program_args(self):
        return ['chmod', '700', self.output().path]

class TwoBitToFa(ExternalProgramTask):
    """Task user for Converting 2bit files to the fa format.

    The task will use a local executable or require the task for obtaining it, and 
    use with the reference genome.

    Parameters:
        ref_url (str): Url for the resource with the reference genome.
        reference_local_file (str): Path for the reference genome 2bit file. If given
            the ``ref_url`` parameter will be ignored.

    Output:
        A `luigi.LocalTarget` for the reference genome fa file.

    """

    ref_url = luigi.Parameter()
    reference_local_file = luigi.Parameter(default='')

    def requires(self):
        if self.reference_local_file != '':
            dependencies = {
                'program': utils.GetProgram(), 
                'file': utils.LocalFile(file=self.reference_local_file),
            }
        else:
            file = path.join(utils.utils.GlobalParams().base_dir, utils.utils.GlobalParams().exp_name+'.2bit')
            dependencies = {
                'program': GetProgram(), 
                'file': utils.Wget(url=self.ref_url, output_file=file),
            }

        return dependencies 

    def output(self):
        return luigi.LocalTarget(path.join(utils.utils.GlobalParams().base_dir,utils.utils.GlobalParams().exp_name+'.fa'))

    def program_args(self):
        return [self.input()['program'].path, self.input()['file'].path, self.output().path]

class GetReferenceFa(utils.MetaOutputHandler, luigi.WrapperTask):
    """Task user for obtaining the reference genome .fa file.

    This task will retrieve an external genome or use a provided local one, and convert
    it from 2bit format to .fa if neccessary.

    Parameters:
        ref_url (str): Url for the resource with the reference genome.
        reference_local_file (str): Path for the reference genome 2bit file. If given
            the ``ref_url`` parameter will be ignored.
        from2bit (bool): Non case sensitive boolean indicating wether the reference genome
            if in 2bit format. Defaults to false.

    Output:
        A `luigi.LocalTarget` for the reference genome fa file.  
    
    """

    reference_local_file = luigi.Parameter(default='')
    ref_url = luigi.Parameter(default='')
    from2bit = luigi.Parameter(default='false')

    def requires(self):
        if self.from2bit.lower() == 'true':
            dependency = TwoBitToFa(reference_local_file=self.reference_local_file ,ref_url=self.ref_url) 
        else:
            if self.reference_local_file != '':
                dependency = utils.LocalFile(self.reference_local_file)
            else:
                out_file = luigi.LocalTarget(path.join(utils.utils.GlobalParams().base_dir, utils.utils.GlobalParams().exp_name+'.fa'))
                dependency = Wget(url=self.ref_url, output_file=out_file)

        return dependency

class FaidxIndex(ExternalProgramTask):
    """Task user for indexing the reference genome .fa file with the samtools faidx utility.

    Aligning the reference genome helps reducing access time drastically.

    Parameters:
        None

    Output:
        A `luigi.LocalTarget` for the .fai index file for the reference genome .  
    
    """

    def requires(self): 
        return GetReferenceFa(from2bit=ReferenceGenome().from_2bit ,ref_url=ReferenceGenome().ref_url)

    def output(self):
        return luigi.LocalTarget(self.input().path+'.fai')

    def program_args(self):
        return ['samtools', 'faidx', self.input()['fa'].path]

class BwaIndex(ExternalProgramTask):
    """Task user for indexing the reference genome .fa file with the bwa index utility.

    Aligning the reference genome helps reducing access time drastically.

    Parameters:
        None

    Output:
        A set of five files are result of indexing the reference genome. The extensions
        for each of the files are '.amb', '.ann', '.bwt', '.pac', '.sa'.
    
    """

    def requires(self): 
        return GetReferenceFa(from2bit=ReferenceGenome().from_2bit ,ref_url=ReferenceGenome().ref_url)

    def output(self):
        outputs = set()

        outputs.add(
            luigi.LocalTarget(
                path.join(utils.GlobalParams().base_dir, utils.GlobalParams().exp_name+".fa.amb")))
        outputs.add(
            luigi.LocalTarget(
                path.join(utils.GlobalParams().base_dir, utils.GlobalParams().exp_name+".fa.ann")))
        outputs.add(
            luigi.LocalTarget(
                path.join(utils.GlobalParams().base_dir, utils.GlobalParams().exp_name+".fa.bwt")))
        outputs.add(
            luigi.LocalTarget(
                path.join(utils.GlobalParams().base_dir, utils.GlobalParams().exp_name+".fa.pac")))
        outputs.add(
            luigi.LocalTarget(
                path.join(utils.GlobalParams().base_dir, utils.GlobalParams().exp_name+".fa.sa")))

        return outputs

    def program_args(self):
        return ['bwa', 'index', self.input().path]

class ReferenceGenome(utils.MetaOutputHandler, luigi.WrapperTask):
    """Higher level task for retrieving the reference genome.
    
    It is given preference to local files over downloading the reference. However the 
    indexing of the reference genome is always done using ``GloablParams.exp_name`` and
    ``GlobalParams.base_dir`` for determining filenames and location for newer files
    respectively.

    The indexing is done using both Samtools and Bwa toolkits.

    Parameters:
        reference_local_file (str) : Optional string indicating the location for the reference genome. If set, it will not be downloaded.
        ref_url (str) : Url for the download of the reference genome.
        from2bit (bool) : A boolean [True, False] indicating whether the reference genome must be converted from 2bit.


    Output:
        A dict mapping keys to `luigi.LocalTarget` instances for each of the processed files. 
        The following keys are available:

        'faidx' : Local file with the index, result of indexing with Samtools.     
        'bwa' : Set of five files, result of indexing the reference genome with Bwa.
        'fa' : Local file with the reference genome.

    """

    reference_local_file = luigi.Parameter(default='',description='Optional string indicating the location for the reference genome. If set, it will not be downloaded.')
    ref_url = luigi.Parameter(default='', description="Url for the download of the reference genome.")
    from2bit = luigi.Parameter(default='false', description="A boolean [True, False] indicating whether the reference genome must be converted from 2bit. Defaultsto false.")

    def requires(self):
        return {
            'faidx' : FaidxIndex(reference_local_file=self.reference_local_file, from2bit=self.from2bit ,ref_url=self.ref_url), \
            'bwa' : BwaIndex(reference_local_file=self.reference_local_file, from2bit=self.from2bit ,ref_url=self.ref_url), \
            'fa' : GetReferenceFa(reference_local_file=self.reference_local_file, from2bit=self.from2bit ,ref_url=self.ref_url) \
            }

if __name__ == '__main__':
    luigi.run(['ReferenceGenome', 
            '--ReferenceGenome-ref-url', 'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/bigZips/hg19.2bit',
            '--ReferenceGenome-from2bit', 'True',
            '--utils.GlobalParams-base-dir', path.abspath(path.curdir),
            '--utils.GlobalParams-log-dir', path.abspath(path.curdir),
            '--utils.GlobalParams-exp-name', 'hg19'])



import luigi
from luigi.contrib.external_program import ExternalProgramTask
from os import path
from wespipeline import utils

from wespipeline.reference import ReferenceGenome
from wespipeline.processalign import AlignProcessing

class PlatypusCallVariants(ExternalProgramTask):
    """Task used for identifying varinats in the bam file provided using Platypus.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the vcf produced.

    Parameters:
        none

    Dependencies:
        ReferenceGenome
        AlignProcessing

    Output:
        A `luigi.LocalTarget` instance for the index vcf file.

    """
    
    def requires(self):
        return {
            'reference' : ReferenceGenome(),
            'process' : AlignProcessing()
        }

    def output(self):
        return luigi.LocalTarget(
            path.join(utils.GlobalParams().base_dir,
            utils.GlobalParams().exp_name+'_platypus.vcf'))

    def program_args(self):
        return ['platypus', 'callVariants', 
            '--bamFiles='+self.input()['process']['bamNoDup']['bam'].path, 
            '--refFile='+self.input()['reference']['fa']['fa'].path, 
            '--output='+self.output().path,
        ]

class FreebayesCallVariants(ExternalProgramTask):
    """Task used for identifying varinats in the bam file provided using Freebayes.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the vcf produced.

    Parameters:
        none

    Dependencies:
        ReferenceGenome
        AlignProcessing

    Output:
        A `luigi.LocalTarget` instance for the index vcf file.

    """
    
    def requires(self):
        return {
            'reference' : ReferenceGenome(),
            'process' : AlignProcessing()
        }

    def output(self):
        return luigi.LocalTarget(
            path.join(utils.GlobalParams().base_dir,
            utils.GlobalParams().exp_name+'_freebayes.vcf'))

    def program_args(self):
        return ['freebayes', 
            '-f', self.input()['reference']['fa']['fa'].path,
            '--bam', self.input()['process']['bamNoDup']['bam'].path,
            '--vcf', self.output().path
        ]

class SamtoolsCallVariants(ExternalProgramTask):
    """Task used for identifying varinats in the bam file provided using Samtools.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the vcf produced.

    Parameters:
        none

    Dependencies:
        ReferenceGenome
        AlignProcessing

    Output:
        A `luigi.LocalTarget` instance for the index vcf file.

    """
    
    def requires(self):
        return {
            'reference' : ReferenceGenome(),
            'process' : AlignProcessing()
        }

    def output(self):
        return luigi.LocalTarget(
            path.join(utils.GlobalParams().base_dir,
            utils.GlobalParams().exp_name+'_samtools.vcf'))

    def program_args(self):
        return [
            self.input()['reference']['fa']['fa'].path,
            self.input()['process']['bamNoDup']['bam'].path,
            self.output().path
        ]

class GatkCallVariants(ExternalProgramTask):
    """Task used for identifying varinats in the bam file provided using GatkCallVariants.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the vcf produced.

    Parameters:
        none

    Dependencies:
        ReferenceGenome
        AlignProcessing

    Output:
        A `luigi.LocalTarget` instance for the index vcf file.

    """
    
    def requires(self):
        return {
            'reference' : ReferenceGenome(),
            'process' : AlignProcessing()
        }

    def output(self):
        return luigi.LocalTarget(
            path.join(utils.GlobalParams().base_dir,
            utils.GlobalParams().exp_name+'_gatk.vcf'))

    def program_args(self):
        return [
            self.input()['reference']['fa']['fa'].path,
            self.input()['process']['bamNoDup']['bam'].path,
            self.output().path
        ]

class DeepcallingCallVariants(ExternalProgramTask):
    """Task used for identifying varinats in the bam file provided using DeepCalling.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the vcf produced.

    Parameters:
        none

    Dependencies:
        ReferenceGenome
        AlignProcessing

    Output:
        A `luigi.LocalTarget` instance for the index vcf file.

    """

    def requires(self):
        return {
            'reference' : ReferenceGenome(),
            'process' : AlignProcessing()
        }

    def output(self):
        return luigi.LocalTarget(
            path.join(utils.GlobalParams().base_dir,
            utils.GlobalParams().exp_name+'_deepcalling.vcf'))

    def program_args(self):
        return [
            self.input()['reference']['fa']['fa'].path,
            self.input()['process']['bamNoDup']['bam'].path,
            self.output().path
        ]

class VariantCalling(utils.MetaOutputHandler, luigi.WrapperTask):
    """Higher level task for the alignment of fastq files.
    
    It is given preference to local files over processing the alignment
    in order to reduce computational overhead. 

    If the bam and bai local files are set, they will be used instead of
    the 

    Alignment is done with the Bwa mem utility.

    Parameters:
        use_platypus (bool) : A non-case sensitive boolean indicating wether to use Platypus for variant callign.
        use_freebayes (bool) : A non-case sensitive boolean indicating wether to use Freebayesfor variant callign.
        use_samtools (bool) : A non-case sensitive boolean indicating wether to use Samtools for variant callign.
        use_gatk (bool) : A non-case sensitive boolean indicating wether to use Gatk for variant callign.
        use_deepcalling (bool) : A non-case sensitive boolean indicating wether to use DeepCalling for variant callign.
        vcf_local_files (string) : A comma delimited list of vfc files to be used instead of using the variant calling tools.

    Output:
        A dict mapping keys to `luigi.LocalTarget` instances for each of the 
        processed files. 
        The following keys are available:
    
        'platypus' : Local file with the variant calls obtained using Platypus.
        'freebayes' : Local file with the variant calls obtained using Freevayes.
        'samtools' : Local sorted file with variant calls obtained using Samtools.
        'gatk' : Local file with the variant calls obtained using GATK.
        'deepcalling' : Local file with the variant calls obtained using DeepCalling.

    """

    use_platypus = luigi.Parameter(default="False", description="A boolean [True, False] indicating wether to create a vcf file using Platypus")
    use_freebayes = luigi.Parameter(default="False",description="A boolean [True, False] indicating wether to create a vcf file using Freebayes")
    use_samtools = luigi.Parameter(default="False",  description="A boolean [True, False] indicating wether to create a vcf file using Samtools")
    use_gatk = luigi.Parameter(default="False", description="A boolean [True, False] indicating wether to create a vcf file using GATK")
    use_deepcalling = luigi.Parameter(default="False", description="A boolean [True, False] indicating wether to create a vcf file using DeepCalling")
    vcf_local_files = luigi.Parameter(default='', description="Comma separated list for the local files to be used instead of processing the alignments.")

    def requires(self):
        dependencies = dict()
        
        if '.vcf' in self.vcf_local_files :
            for file in self.vcf_local_files.split(','):
                dependencies.update({file.strip()})
        else:
            if self.use_platypus == 'true':
                dependencies.update({'platypus': PlatypusCallVariants()})

            if self.use_freebayes == 'true':
                dependencies.update({'freebayes': FreebayesCallVariants()})

            if self.use_samtools == 'true':
                dependencies.update({'samtools': SamtoolsCallVariants()})

            if self.use_gatk == 'true':
                dependencies.update({'gatk': GatkCallVariants()})

            if self.use_deepcalling == 'true':
                dependencies.update({'deepcalling': DeepcallingCallVariants()})

        return dependencies


if __name__ == '__main__':
    luigi.run(['VariantCalling', 
            '--VariantCalling-use-platypus', 'true',
            '--VariantCalling-use-freebayes', 'true',
            '--VariantCalling-use-samtools', 'false',
            '--VariantCalling-use-gatk', 'false',
            '--VariantCalling-use-deepcalling', 'false',
            '--AlignProcessing-cpus', '6',
            '--FastqAlign-cpus', '6', 
            '--FastqAlign-create-report', 'True', 
            '--GetFastq-fastq1-url', '',
            '--GetFastq-fastq2-url', '',
            '--GetFastq-from-ebi', 'False',
            '--GetFastq-paired-end', 'True',
            '--ReferenceGenome-ref-url', 'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/bigZips/hg19.2bit',
            '--ReferenceGenome-from2bit', 'True',
            '--utils.GlobalParams-base-dir', './experiment',
            '--utils.GlobalParams-log-dir', path.abspath(path.curdir),
            '--utils.GlobalParams-exp-name', 'hg19'])



import luigi
from luigi.contrib.external_program import ExternalProgramTask
from os import path
from wespipeline import utils

from wespipeline.reference import ReferenceGenome
from wespipeline.align import FastqAlign

class SortSam(ExternalProgramTask):
    """Task used for sorting the alignment sam file.

    It requires the output of the ``wespipeline.reference.FastqAlign`` step.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the Bam file produced.

    Parameters:
        none

    Output:
        A `luigi.LocalTarget` instance for the sorted Sam Bam file.

    """

    def requires(self):
        return FastqAlign()

    def output(self):
        return luigi.LocalTarget(
            path.join(utils.GlobalParams().base_dir,
            utils.GlobalParams().exp_name+'.bam')
            )

    def program_args(self):
        return ['samtools', 'sort', '-@', AlignProcessing().cpus, 
            self.input()['sam'].path, '-o', self.output().path
        ]

class IndexBam(ExternalProgramTask):
    """Task used for indexing the Bam file.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the Bai file produced.

    Parameters:
        none

    Output:
        A `luigi.LocalTarget` instance for the index Bai file.

    """

    def requires(self):
        return SortSam()

    def output(self):
        return luigi.LocalTarget(self.input().path+'.bai')

    def program_args(self):
        return ['samtools', 'index', '-@', AlignProcessing().cpus, self.input().path]

class PicardMarkDuplicates(ExternalProgramTask):
    """Task used for removing duplicates from the Bam file.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the Bam file produced.

    Parameters:
        none

    Output:
        A `luigi.LocalTarget` instance for the Bam file without the duplicates.

    """

    def requires(self):
        return {'index' : IndexBam(),
            'bam' : SortSam()
        }

    def output(self):
        return {
            'bam' : luigi.LocalTarget( \
                path.join(utils.GlobalParams().base_dir, utils.GlobalParams().exp_name+'_nodup.bam')),
            'metrics' : luigi.LocalTarget( \
                path.join(utils.GlobalParams().base_dir, utils.GlobalParams().exp_name+'_MD.matrix'))
        }

    def program_args(self):
        return ['picard', 'MarkDuplicates', 
            'I='+self.input()['bam'].path, 
            'O='+self.output()['bam'].path, 
            'METRICS_FILE='+self.output()['metrics'].path,
            'REMOVE_DUPLICATES=true'
        ]

class IndexNoDup(ExternalProgramTask):
    """Task used for indexing the Bam file without duplicates.

    The ``wespipeline.utils.GlobalParams.exp_name`` will be used for giving name
    to the Bai file produced.

    Parameters:
        none

    Output:
        A `luigi.LocalTarget` instance for the index Bai file.

    """

    def requires(self):
        return PicardMarkDuplicates()

    def output(self):
        return luigi.LocalTarget(self.input()['bam'].path+'.bai')

    def program_args(self):
        return ['samtools', 'index', '-@', AlignProcessing().cpus, self.input()['bam'].path]

class AlignProcessing(utils.MetaOutputHandler, luigi.WrapperTask):
    """Higher level task for the alignment of fastq files.
    
    It is given preference to local files over processing the alignment
    in order to reduce computational overhead. 

    If the bam and bai local files are set, they will be used instead of
    the 

    Alignment is done with the Bwa mem utility.

    Parameters:
        bam_local_file (str) : String indicating the location of a local
            bam file with the sorted alignment. If set, this file will not be created.
        bai_local_file (str) : String indicating the location of a local
            bai file with the index for the alignment. If set, this file will not be created.
        no_dup_bam_local_file (str) : String indicating the location of a local
            sam file without the duplicates. If set, this file will not be created.
        no_dup_bai_local_file (str) : String indicating the location of a local
            file with the index for the bam file without duplicates. If set, this 
            file will not be created.
        cpus (int): Integer indicating the number of cpus that can be used for 
            the alignment.

    Output:
        A dict mapping keys to `luigi.LocalTarget` instances for each of the 
        processed files. 
        The following keys are available:
    
        'bam' : Local file with the sorted alignment.
        'bai' : Local file with the alignment index.
        'bamNoDup' : Local sorted file with duplicates removed.
        'indexNoDup' : Local file with the index for sorted alignment without duplicates.

    """

    cpus = luigi.Parameter(description="Number of cpus to be used by each task thread.")
    bam_local_file = luigi.Parameter(default='', description='Optional path for the file. If set, wil be skipped.')
    bai_local_file = luigi.Parameter(default='', description='Optional path for the file. If set, wil be skipped.')
    no_dup_bam_local_file = luigi.Parameter(default='', description='Optional path for the file. If set, wil be skipped.')
    no_dup_bai_local_file = luigi.Parameter(default='', description='Optional path for the file. If set, wil be skipped.')

    def requires(self):
        if self.no_dup_bam_local_file != '' and self.no_dup_bai_local_file != '':
            bam = utils.LocalFile(file_path=self.no_dup_bam_local_file)
            bai = utils.LocalFile(file_path=self.no_dup_bai_local_file)
            bamNoDup = utils.LocalFile(file_path=self.no_dup_bam_local_file)
            baiNoDup = utils.LocalFile(file_path=self.no_dup_bai_local_file)
        else:
            bamNoDup = Picardmarkduplicates()
            baiNoDup = IndexBam()

        if self.bam_local_file != '' and self.bai_local_file != '':
            bam = utils.LocalFile(file_path=self.bam_local_file)
            bai = utils.LocalFile(file_path=self.bai_local_file)
        else:
            bam = SortSam()
            bai = IndexSam()

        return {
            'bam' : bam,
            'bai' : bai,
            'bamNoDup' : bamNoDup,
            'indexNoDup' : baiNoDup
            }

if __name__ == '__main__':
    luigi.run(['AlignProcessing', 
            '--AlignProcessing-cpus', '6',
            '--FastqAlign-cpus', '6', 
            '--FastqAlign-create-report', 'True', 
            '--GetFastq-fastq1-url', '',
            '--GetFastq-fastq2-url', '',
            '--GetFastq-from-ebi', 'False',
            '--GetFastq-paired-end', 'True',
            '--ReferenceGenome-ref-url', 'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/bigZips/hg19.2bit',
            '--ReferenceGenome-from2bit', 'True',
            '--utils.GlobalParams-base-dir', path.abspath(path.curdir),
            '--utils.GlobalParams-log-dir', path.abspath(path.curdir),
            '--utils.GlobalParams-exp-name', 'hg19'])


import luigi
from luigi.contrib.external_program import ExternalProgramTask
from os import path
from wespipeline import utils 

from wespipeline.reference import ReferenceGenome
from wespipeline.fastq import GetFastq

class BwaAlignFastq(ExternalProgramTask):
    """Task used for aligning fastq files against the reference genome.

    It requires the output of both the wespipeline.reference.ReferenceGenome and
    wespipeline.fastq.GetFastq higher level tasks in order to proceed with 
    the alignment.

    If ``wespipeline.utils.GlobalParams.exp_name`` is set, it will be used for giving name
    to the Sam file produced.

    Parameters:
        none

    Output:
        A `luigi.LocalTarget` instance for the aligned sam file.

    """

    def requires(self): 
        return {
            'reference' : ReferenceGenome(),
            'fastq' : GetFastq()
        }

    def output(self):
        get_name = lambda x: os.path.splitext(os.path.basename(os.path.realpath(x)))[0]
        output_filename = get_name(self.input['fastq']['fastq1'].path) + '.sam'

        if utils.GlobalParams().exp_name:
            output_filename = path.join(utils.GlobalParams().base_dir, utils.GlobalParams().exp_name+".sam")
        
        return luigi.LocalTarget(output_filename)

    def program_args(self):
        name = 'wespipeline' if not utils.GlobalParams().exp_name else utils.GlobalParams().exp_name

        rg = '@RG\\tID:'+name+'\\tSM:'+name+'\\tPL:illumina'

        args = ['bwa', 'mem', '-M', '-R', rg,
            '-t', FastqAlign().cpus,
            self.input()['reference']['fa'].path,
            self.input()['fastq']['fastq1'].path,
            ]

        if 'fastq_2' in self.input()['fastq']:
            args.append(self.input()['fastq']['fastq2'].path)

        args.append('-o')
        args.append(self.output().path)

        return args

class FastqAlign(utils.MetaOutputHandler, luigi.WrapperTask):
    """Higher level task for the alignment of fastq files.
    
    It is given preference to local files over processing the alignment
    in order to reduce computational overhead. 

    Alignment is done with the Bwa mem utility.

    Parameters:
        fastq1_local_file (str): String indicating the location of a local
            Sam file for the alignment. 
        cpus (int): Integer indicating the number of cpus that can be used for 
            the alignment.

    Output:
        A dict mapping keys to `luigi.LocalTarget` instances for each of the 
        processed files. 
        The following keys are available:

        'sam' : Local file with the alignment.     

    """

    sam_local_file = luigi.Parameter(default='', description='Optional file path for the aligned sam file. If set, the alignment will be skipped.')
    cpus = luigi.Parameter(default='', description="Number of cpus to be used by each task thread.")

    def requires(self):
        if self.sam_local_file != '':
            return {'sam': utils.LocalFile(file_path=self.sam_local_file)}
        else:
            return {'sam' : BwaAlignFastq()}

if __name__ == '__main__':
    luigi.run(['FastqAlign', 
            '--FastqAlign-create-report', 'False',
            '--FastqAlign-cpus', '8',
            '--GetFastq-fastq1-url', '',
            '--GetFastq-fastq2-url', '',
            '--GetFastq-from-ebi', 'False',
            '--GetFastq-paired-end', 'False',
            '--ReferenceGenome-ref-url', 'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/bigZips/hg19.2bit',
            '--ReferenceGenome-from2bit', 'True',
            '--utils.GlobalParams-base-dir', path.abspath('./experiment'),
            '--utils.GlobalParams-log-dir', path.abspath(path.curdir),
            '--utils.GlobalParams-exp-name', 'hg19'])



import os
import luigi
from luigi.contrib.external_program import ExternalProgramTask

class GlobalParams(luigi.Config):
    """Task used for specifying globally accessible parameters.

    Parameters defined in this class are task independent and should
    mantain low.

    Parameters:
        exp_name (str): Name for the experiment. Useful for defining file names.
        log_dir (str): Absolute path for the logs of the application.
        base_dir (str): Absolute path to the directory where files are expected
            to appear if not specifyed differently.

    """

    exp_name = luigi.Parameter(description='(str): Name for the experiment. Useful for defining file names.')
    log_dir = luigi.Parameter(description='(str): Absolute path for the logs of the application.')
    base_dir = luigi.Parameter(description='(str): Absolute path to the directory where files are expected' \
        + 'to appear if not specifyed differently.')

class MetaOutputHandler:
    """Helper class for propagating inputs in WrapperTasks"""

    def output(self):
        if type(self.input()) == luigi.LocalTarget:
            return self.input()
        elif type(self.input()) == dict:
            return {key: task for key,task in self.input().items()}

class LocalFile(luigi.Task):
    """Helper task for making.

    No extra processing is done in the task. It allows to make tasks
    dependent on files using the same strategy as with other tasks.

    Parameters:
        file (str): Absolute path to the file to be tested.
    """

    file = luigi.Parameter(description='(str): Absolute path to the file to be tested.')

    def output(self):
        return luigi.LocalTarget(self.file)

    def run(self):
        if not os.path.isfile(self.file):
            raise Exception(f'{self.file} could not be found.')

class Wget(ExternalProgramTask):
    """Task for downloading files using the tool wget.

    Parameters:
        url (str): Url indicating the location of the resource to be retreived.
        output_file (str): Absolute path for the destiny location of the retrived resource.
    """

    url = luigi.Parameter(description='(str): Url indicating the location of the resource to be retreived.')
    output_file = luigi.Parameter(description='(str): Absolute path for the destiny location of the retrived resource.')

    def requires(self):
        return None

    def output(self):
        return luigi.LocalTarget(self.output_file)

    def program_args(self):
        return ['wget', '-O', self.output().path, self.url]

class GunzipFile(ExternalProgramTask):
    """Task for unzipping compressed files.

    Gunzip will allways do the process inplace, deleting the extension.

    Parameters:
        input_file (str): Absolute path to the compressed file.
    """

    input_file = luigi.Parameter(description='(str): Absolute path to the compressed file.')

    def requires(self):
        return None

    def output(self):
        return luigi.LocalTarget('.'.join(self.input_file.split('.')[:-1]))

    def program_args(self):
        if os.path.isfile(self.input_file):
            return ['gunzip', '-d', self.input_file]
        else:
            raise Exception(f"Can't uncompress {self.input_file}. The file was not found")

class UncompressFile(ExternalProgramTask):
    """Task for unzipping compressed files to a desired location.

    Gunzip will allways do the process inplace, deleting the extension. This
    task allows to select the destination.

    This operation

    Parameters:
        input_file (str): Absolute path to the compressed file.
        output_file (str): Absolute path to the desired final location.
        copy (bool): Non case sensitive boolean indicating wether to copy or
            to move the file. Defaults to false.
    """

    input_file = luigi.Parameter(description='(str): Absolute path to the compressed file.')
    output_file = luigi.Parameter(description='(str): Absolute path to the'
        +' desired final location.')
    copy = luigi.Parameter(default='false', description='(bool): Non case sensitive'
        + 'boolean indicating wether to copy or to move the file. Defaults to false.')

    def requires(self):
        return GunzipFile(self.input_file)

    def output(self):
        return luigi.LocalTarget(self.output_file)

    def program_args(self):
        if os.path.isfile(self.input().path):
            command = 'cp' if self.copy.lower() == 'true' else 'mv'

            if os.path.abspath(self.input().path) != os.path.abspath(self.output_file):
                return [command, self.input().path, self.output_file]
            else:
                return ['echo', 'Input and output files are the same']
        else:
            raise Exception(f"Can't uncompress {self.input_file}. The file was not found")



import os
import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wespipeline",
    version="0.9.0",
    author="Alejandro Rodríguez Díaz",
    author_email="jancho@usal.es",
    description="An implementation of a whole exome analysis pipeline using the library Luigi for workflow management.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    project_urls={
        'Documentation': 'https://wespipeline.readthedocs.io/en/latest/index.html',
        'Source': 'https://github.com/Janchorizo/wespipeline',
        'Tracker': 'https://github.com/Janchorizo/wespipeline/issues',
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'luigi',
        'python-daemon'
    ]
)



import itertools
import luigi
from luigi.contrib.external_program import ExternalProgramTask
from os import path
from wespipeline import utils 

from wespipeline.vcf import VariantCalling

class VcftoolsCompare(ExternalProgramTask):
    """Task used for comparing a pair of vcf files using VcfTools.

    Parameters:
        vcf1 (str): Absolute path to the first file to be compared.
        vcf2 (str): Absolute path to the second file to be compared.

    Dependencies:
        None

    Output:
        A `luigi.LocalTarget` instance for the result of comparing the files.

    """

    vcf1 = luigi.Parameter()
    vcf2 = luigi.Parameter()

    def requires(self):
        return None

    def output(self):
        filename = lambda f: path.split(f)[-1].split('.')[0]

        return luigi.LocalTarget(
            path.join(utils.GlobalParams().base_dir,
            ''.join([filename(self.vcf1), '_vs_', filename(self.vcf2)])))

    def program_args(self):
        return ['vcftools',
            '--vcf',
            self.vcf1,
            '--diff',
            self.vcf2,
            '--not-chr',
            '--diff-site',
            '--out',
            self.output().path
        ]

class VcftoolsAnalysis(ExternalProgramTask):
    """Task used for extracting basic statistics for the variant calls using VcfTools.

    Parameters:
        vcf (str): Absolute path to the file with the variant annotations.

    Dependencies:
        None

    Output:
        A `luigi.LocalTarget` instance for the file with the vcf statistics.

    """

    vcf = luigi.Parameter()

    def requires(self):
        return None

    def output(self):
        return luigi.LocalTarget(self.vcf.replace('.vcf', '.vcf_info'))

    def program_args(self):
        return ['vcftools',
            '--vcf',
            self.vcf,
            '--out',
            self.output().path
        ]

class VariantCallingAnalysis(luigi.Task):
    """Higher level task for comparing variant calls.
    
    Comparing variant calls is a delicate task that increments in complexity when
    dealing in diploid sequences (such us the human genome), where different variants
    can appear in the same position in each of the pair chromomes.

    The normalization is done with vt, and the comparison with VcfTools

    Parameters:
        None

    Output:
        None. The resulting files are not provided as task output. Each of the n vcf files is analyzed and comparied by pairs. 
        It is a total of 2n-1 files.

    """

    def requires(self):
        return VariantCalling()

    def outputs(self):
        output = [vcf.path.replace('.vcf','vcf_info') for vcf in self.input().values()]

        return output

    def run(self):
        yield((VcftoolsAnalysis(vcf=vcf.path) for vcf in self.input().values()))
        if len(self.input()) > 1:
            yield((VcftoolsCompare(vcf1=vcf1.path, vcf2=vcf2.path) \
                for vcf1,vcf2 in itertools.combinations(self.input().values(),2)))
    

if __name__ == '__main__':
    luigi.run(['VariantCallingAnalysis', 
            '--VariantCalling-use-platypus', 'true',
            '--VariantCalling-use-freebayes', 'true',
            '--VariantCalling-use-samtools', 'false',
            '--VariantCalling-use-gatk', 'false',
            '--VariantCalling-use-deepcalling', 'false',
            '--AlignProcessing-cpus', '6',
            '--FastqAlign-cpus', '6', 
            '--FastqAlign-create-report', 'True', 
            '--GetFastq-fastq1-url', '',
            '--GetFastq-fastq2-url', '',
            '--GetFastq-from-ebi', 'False',
            '--GetFastq-paired-end', 'True',
            '--ReferenceGenome-ref-url', 'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/bigZips/hg19.2bit',
            '--ReferenceGenome-from2bit', 'True',
            '--utils.GlobalParams-base-dir', './experiment',
            '--utils.GlobalParams-log-dir', path.abspath(path.curdir),
            '--utils.GlobalParams-exp-name', 'hg19'])
